"""
----------
 Examples
----------

Get all documents in an RSpace Folder or Notebook:

.. image:: images/RSpace_UV-vis.png
    :alt: Notebook contents.
    :align: center

.. code-block:: python

    import inm_rspace as rs
    docs = rs.get_docs_in_notebook(7074)
    print([doc['name'] for doc in docs])

Output:

.. image:: images/Code_UV-vis.png
    :alt: The generated output.
    :align: center

-------------------
 API documentation
-------------------

"""

import os
from xml.dom.minidom import parseString as parse_xml
from fnmatch import fnmatch
from rspace_client.eln import eln
from rspace_client.inv import inv

class ELNDummy:
  """Dummy ELN object for testing
  """
  
  def __init__(self):
    return
  
  def list_folder_tree(*args):
    return {'records': []}
  
  def get_document(*args):
    return {'name': '', 'form': dict()}

  def get_folder(*args):
    return {'name': ''}

  def get_forms(*args):
    return {'forms': []}

  def get_form(*args):
    return {'name': '', 'fields': []}

  def create_form(self, name, fields):
    return {'name': name, 'fields': fields}

  def publish_form(self, globalId):
    return


try:
  ELN = eln.ELNClient(os.getenv("RSPACE_URL"), os.getenv("RSPACE_API_KEY"))
  Inventory = inv.InventoryClient(os.getenv("RSPACE_URL"), os.getenv("RSPACE_API_KEY"))
except:
  ELN = ELNDummy()
  Inventory = ELNDummy()

replace = {' ': '_', ',': '.', '<p>': '', '</p>': ''}



def html_ref(rspace_obj):
  """html string to reference an Rspace object
  
  Parameters
  ----------
  rspace_obj : file or document
      Rspace object
  
  Returns
  -------
  string : str
      string that can be inserted in an html string to reference the given object.
  
  Raises
  ------
  ValueError
      raised if the rspace object type is not recognized
  """
  link = rspace_obj['_links'][0]['link']
  obj_loc = link.split('/')[-2]
  if obj_loc=='documents':
    id_str = 'docId'
  elif obj_loc=='files':
    id_str = 'fileId'
  else: 
    raise ValueError(f'Unknown rspace object type: {obj_loc}')
  
  return f"<{id_str}={rspace_obj['id']}>"

def get_line(string, index):
  idx_beg = idx_end = index
  while string[idx_beg]!='\n': idx_beg -= 1
  while string[idx_end]!='\n': idx_end += 1

  return string[(idx_beg+1):idx_end]



def tables_from_xml(xml_string, file, delimiter=',', replace=replace):
  """
  extract all tabular data from an xml string and save it as a csv file.
  
  Parameters
  ----------
  xml_string : str
      input xml string
  file : str
      file path to use, although an appendix is going to be inserted 
      to enumerate multiple tables in the xml string.
  delimiter : str, optional
      field delimiter to be used in the csv file
  replace : dict, optional
      key,value pairs indicating strings (keys) to be replaced with their corresponding value.
  
  Returns
  -------
  files : list<str>
      List of files exported.
  """

  num = 0 # used to name file exports
  stub, ext = os.path.splitext(file)
  if ext=='': ext = '.csv'

  files = []
  while True:
    # any table is delimited by the tag 'tbody'
    idx_tab_beg = xml_string.find('<tbody')
    if idx_tab_beg < 0: break
    idx_tab_end = xml_string.find('</tbody>')+8

    # some replacements are necessary for successful parsing
    table_str = xml_string[idx_tab_beg:idx_tab_end]
    table_str = table_str.replace('\n','')
    table_str = table_str.replace('""','\'')
    table_str = table_str.replace('&nbsp;',' ')

    # parse table and save as rows, as indicated by the tag 'tr'
    table_xml = parse_xml(table_str)
    rows = table_xml.getElementsByTagName('tr')
    
    # write the detected table to a file
    num += 1
    outfile = f'{stub}_{str(num).zfill(2)}{ext}'
    files.append(outfile)
    with open(outfile,'w') as fid:
      for row in rows:
        for element in row.getElementsByTagName('td'):
          try: 
            # cells containing text with formatting of tag 'span'
            sub_elements = element.getElementsByTagName('span')
            value = sub_elements[0].firstChild.nodeValue
          except: 
            try:
              # cells containing text with formatting of tag 'p'
              sub_elements = element.getElementsByTagName('p')
              value = sub_elements[0].firstChild.nodeValue
            except:
              # cells containing unformatted text
              value = element.firstChild.nodeValue

          value = str(value)
          for char in replace.keys():
            value = value.replace(char, replace[char])

          fid.write(value+delimiter)
        fid.write('\n')

    # remove the processed part from the xml_string before the next iteration
    xml_string = xml_string[idx_tab_end:]

  return files



def get_files(document, field_key=None):
  """list files attached to (a field in) an Rspace document
  
  Parameters
  ----------
  document : RspaceDocument
      input document
  field_key : int or str, optional
      name of the field, from which files are extracted. 
      If None, files from all fields are listed.
  
  Returns
  -------
  files : list<tuple<str,str>>
      files found as globalId,filename-pairs.
  """
  if isinstance(field_key, str): 
    fields = [f['name'] for f in document['fields']]
    field_key = fields.index(field_key)
  elif field_key is None:
    files = []
    for ifield in range(len(document['fields'])):
      files += get_files(document, ifield)
    return files
  
  return document['fields'][field_key]['files']
  
def get_docs_in_notebook(notebook_id, form_pattern=None, verbose=False):
  """
  scan for Rspace documents in a given folder whose form name matches a pattern
  
  Parameters
  ----------
  notebook_id : str
      notebookID of the Rspace notebook to search for matches
  form_pattern : str
      glob-style pattern that the form name must match

  Returns
  -------
  results : list<dict>
      list of documents matching the form name
  """

  results = []
  records = ELN.list_folder_tree(notebook_id)
  nb_name = ELN.get_folder(notebook_id)['name']
  for page in records['records']:
    doc = ELN.get_document(page['id'])
    print(f"- {nb_name}/{doc['name']} ({doc['form']['name']})")
    if form_pattern is None: 
      results.append(doc)
      continue
    form_name = doc['form']['name']
    if fnmatch(form_name, form_pattern): results.append(doc)

  return results

def get_docs_in_folder(folder_id, form_pattern=None, verbose=False):
  """
  scan for Rspace documents in a given folder whose form name matches a pattern
  
  Parameters
  ----------
  folder_id : str
      folderID of the Rspace folder to search for matches
  form_pattern : str
      glob-style pattern that the form name must match

  Returns
  -------
  results : list<dict>
      list of documents matching the form name
  """
  records = ELN.list_folder_tree(folder_id)
  results = []
  for share in records['records']:
    if share['type']=='NOTEBOOK':
      results += get_docs_in_notebook(share['id'], form_pattern=form_pattern, verbose=verbose)
      continue
    
    doc = ELN.get_document(share['id'])
    if verbose: print(f"- {doc['name']} ({doc['form']['name']})")
    if form_pattern is None: 
      results.append(doc)
      continue
    form_name = doc['form']['name']
    if fnmatch(form_name, form_pattern): results.append(doc)

  return results


def get_requests(shared_folder_id, verbose=False):
  """get all shared documents requesting a workflow to be performed
  
  Parameters
  ----------
  shared_folder_id : str
      folderId of the "Shared" Folder in Rspace
  
  Returns
  -------
  results : list<dict>
      list of shared Rspace documents using a `Request:*` form
  """
  user_folders = ELN.list_folder_tree(shared_folder_id)
  results = []
  for folder in user_folders['records']:
    if verbose: print(f"{folder['name']} ({folder['id']}):")
    results += get_docs_in_folder(folder['id'], 'Request:*', verbose=verbose)
    if verbose: print()

  return results

def get_field(document, field_name):
  """get (the first) field from an Rspace document dict with a given name.
  
  Parameters
  ----------
  document : dict
      input document
  field_name : str
      name of the field to be accessed
  
  Returns
  -------
  field : dict
      field with the given name
  """
  for field in document['fields']:
    if field['name']==field_name: return field
  return {}

def field_index(document, field_name):
  """get index of the (first) field from an Rspace document dict with a given name.
  
  Parameters
  ----------
  document : dict
      input document
  field_name : str
      name of the field to be accessed
  
  Returns
  -------
  idx : int
      the index of the field with the given name
  """
  for idx, field in enumerate(document['fields']):
    if field['name']==field_name: return idx
  return -1

def compare_forms(form1, form2):
  """determine whether or not two forms have identical names and fields.
  
  Parameters
  ----------
  form1 : dict
      a dict corresponding to an RSpace form containing at least the keys 'name' and 'fields'.
  form2 : dict
      a dict corresponding to an RSpace form containing at least the keys 'name' and 'fields'.

  Returns
  -------
  match : bool
      True if both forms are identical, False otherwise.
  """

  if form1['name'] != form2['name']: return False
  field_names1 = [field['name'] for field in form1['fields']]
  field_names2 = [field['name'] for field in form2['fields']]
  if len(field_names1) != len(field_names2): return False
  field_types1 = [field['type'] for field in form1['fields']]
  field_types2 = [field['type'] for field in form2['fields']]
  if len(field_types1) != len(field_types2): return False

  for val in field_names1:
    if val not in field_names2: return False
  for val in field_types1:
    if val not in field_types2: return False
  
  return True

def get_form_by_dict(new_form):
  """if it exists, get the Rspace form matching a dict. Otherwise, create a new form first.
  
  Parameters
  ----------
  new_form : dict
      a dict corresponding to an RSpace form containing at least the keys 'name' and 'fields'.
  
  Returns
  -------
  rs_form : dict
      the found/newly created RSpace form. 
  """
  forms = ELN.get_forms()['forms']
  found_form = False
  for form in forms:
    form = ELN.get_form(form['id'])
    if compare_forms(new_form, form):
      return form
  
  rs_form = ELN.create_form(new_form['name'], fields=new_form['fields'])
  print(f"Newly created Form: {rs_form['globalId']}")
  ELN.publish_form(rs_form['globalId'])
  return rs_form



"""
----------
 Examples
----------

A fully outomated workflow that takes RSpace documents with attached CSV files and plots them:

The Input fields in the RSpace document look like this:

.. image:: images/Input_PlotColumnsCSV.png
    :alt: Notebook Inputs.
    :align: center


The automatically generated Output field looks like this:

.. image:: images/Output_PlotColumnsCSV.png
    :alt: Automatically generated Notebook Output.
    :align: center

The code for this example can be found `here <https://github.com/sintharic/inm-rspace/tree/main/examples>`_.

-------------------
 API documentation
-------------------

"""

import os, sys
import shutil
import traceback
from pathlib import Path
from datetime import datetime
from fnmatch import fnmatch
from . import core



SHARED_FOLDER_ID = '1018'
HOME = str(Path.home())
ERROR_CODE = {
  "SUCCESS": 0,
  "ALREADY_COMPLETED": 1,
  "WRONG_WORKFLOW": 2,
  "WRONG_FILES": 3,
  "WRONG_KWARGS": 4,
  "FAILED_DOWNLOAD": 5,
  "FAILED_UPLOAD": 6,
  "FAILED_WORKFLOW": 7
}
ERROR_NAME = {v: k for k, v in ERROR_CODE.items()}



def get_file_paths(field):
  """Not yet working!
  """
  
  files = []
  lines = field['content'].split('\n')
  xml = core.parse_xml(field['content'])
  paragraphs = xml.getElementsByTagName('p')
  paragraphs = [str(p) for p in paragraphs]
  for paragraph in paragraphs:
    print(paragraph)
    if os.path.isfile(paragraph): files.append(paragraph)

  return files


class Workflow:
  def __init__(self, document: dict, path=HOME):
    self.name = str(self.__class__.mro()[0]).split('.')[-1][:-2]
    self.document = document
    self.directory = f"{path}{os.sep}{self.name}{os.sep}{self.document['globalId']}_{self.document['name']}"
    self.time_signature = ''

    self.expected = dict()
    self.field_name = {'input': 'Unknown', 'output': 'Unknown', 'workflow': 'Unknown', 'kwargs': 'Unknown', 'completed': 'Unknown'}

    self.description = ''
    self.define()

    self.init_members()

  def define(self):
    """Define the properties of the Workflow.

    This function should be redefined for every subclass of Workflow.
    """
    self.field_name['workflow'] = 'Workflow'
    self.field_name['completed'] = 'Completed'

    self.expected['input'] = ['*']
    self.field_name['input'] = 'Input Data'
    
    self.expected['output'] = ['*.txt']
    self.field_name['output'] = 'Output Data'
    
    self.expected['kwargs'] = []
    self.field_name['kwargs'] = 'Arguments (JSON)'
    
    self.description = ''

  def init_members(self):
    self.time_signature = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    self.info = f"Workflow '{self.name}' initialized from {self.document['globalId']} on {self.time_signature}.\n"
    self.code = 0
    self.traceback = ''
    self.kwargs = dict()
    self.input_files = []
    self.output_files = []

  def prepare(self):
    """Prepare for workflow execution: 

- check if request has already been completed
- check if the requested workflow matches this one
- create a working directory for the workflow
- try to download files from the request
    """
    self.init_members()

    self.check_completed()
    if self.code: return
    
    os.makedirs(self.directory, exist_ok=True)

    self.check_workflow()
    if self.code: return
    
    self.get_args()
    self.get_input_files()
  
  def check_completed(self):
    """Check if the requested workflow has already been completed.
    """
    completed = core.get_field(self.document, self.field_name['completed'])['content']
    if completed != 'no': 
      self.code = ERROR_CODE['ALREADY_COMPLETED']

  def check_workflow(self):
    """Check if the requested workflow matches this one.
    """
    if self.field_name['workflow'] is None: return

    workflow = core.get_field(self.document, self.field_name['workflow'])['content']
    if workflow!=self.name: 
      self.code = ERROR_CODE['WRONG_WORKFLOW']

  def get_args(self):
    """Read the keyword arguments from the kwargs field of the Rspace document.
    """
    # if len(self.expected['kwargs']) == 0: return
    if self.field_name['kwargs'] is None: return

    kwargs = core.get_field(self.document, self.field_name['kwargs'])['content']
    if not isinstance(kwargs, str):
      self.kwargs = dict()
      return
    
    kwargs = kwargs.lstrip(' ')
    kwargs = kwargs.rstrip(' ')
    print('DEBUG', kwargs)#DEBUG
    if kwargs=='':
      self.kwargs = dict()
      return
    if kwargs[0]!='{': kwargs = '{'+kwargs
    if kwargs[-1]!='}': kwargs = kwargs+'}'

    try: self.kwargs = dict(eval(kwargs))
    except:
      self.traceback += traceback.format_exc()
      self.code = ERROR_CODE['WRONG_KWARGS']

  def check_args(self, fun):
    if len(self.kwargs)==0: return
    try:
      kwdefaults = fun.__kwdefaults__
      for key in self.kwargs.keys():
        if key not in kwdefaults.keys():
          raise KeyError(f"Specified keyword '{key}' not in kwargs of function '{fun.__name__}': {kwdefaults}")
    except:
      self.code = ERROR_CODE['WRONG_KWARGS']
      self.traceback += traceback.format_exc()


  def get_input_files(self):
    """Check if all expected input files have been provided in the request. 
    If yes, download them.
    """
    files_found = core.get_files(self.document, self.field_name['input'])
    files_matched = [{'name': '__missing__'}]*len(self.expected['input'])
    for ifile, file_pattern in enumerate(self.expected['input']):
      for file in files_found:
        if fnmatch(file['name'], file_pattern): files_matched[ifile] = file

    for file in files_matched:
      if file['name']=='__missing__': 
        self.code = ERROR_CODE['WRONG_FILES']
        return

    self.download_files(files_matched)

  def download_files(self, files):
    """Download files from the Rspace Gallery into this workflow's working directory.
    
    Parameters
    ----------
    files : list<dict>
        List of Rspace file objects
    """
    for file in files:
      filepath = f"{self.directory}{os.sep}{file['name']}"
      try: core.ELN.download_file(file['id'], filepath)
      except: 
        self.traceback += traceback.format_exc()
        self.code = ERROR_CODE['FAILED_DOWNLOAD']
        return
      self.input_files.append(filepath)

  def workflow(self):
    """The actual workflow to be executed
    """
    if not self.code: 
      self.info += f"Successfully downloaded files: {[file['name'] for file in self.input_files]}\n"
    filepath = f"{self.directory}{os.sep}result.txt"
    with open(filepath, 'w') as fid:
      fid.write('Dummy result of base Workflow class.')
    self.output_files.append(filepath)

  def summary(self):
    """Summarize the success of the workflow based on generated error codes.
    """
    msg = f'Exit with the error code {self.code}: '
    if ERROR_NAME[self.code] == 'ALREADY_COMPLETED':
      msg += 'Request already completed. Nothing to be done.'
    elif ERROR_NAME[self.code] == 'WRONG_WORKFLOW':
      msg += 'Requested workflow does not match.\n'
      workflow = core.get_field(self.document, self.field_name['workflow'])['content']
      msg += f"Initialized workflow is '{self.name}' with workflow field '{self.field_name['workflow']}', but {self.document['globalId']} requests '{workflow}'."
    elif ERROR_NAME[self.code] == 'WRONG_FILES':
      msg += 'Supplied files do not match.\n'
      msg += f"Workflow '{self.name}' expects the following files: {self.expected['input']}"
    elif ERROR_NAME[self.code] == 'WRONG_KWARGS':
      msg += f"Supplied field {self.field_name['kwargs']} is wrong: '{core.get_field(self.document, self.field_name['kwargs'])['content']}'\n"
      msg += f"Workflow '{self.name}' expects the following kwargs: {self.expected['kwargs']}\n"
      msg += 'There may be an traceback attached to this field with more info.'
    elif ERROR_NAME[self.code] == 'FAILED_DOWNLOAD':
      msg += 'Unable to download files.\n'
      msg += 'Check the attached traceback for more info.'
    elif ERROR_NAME[self.code] == 'FAILED_WORKFLOW':
      msg += 'Error while executing the workflow itself.\n'
      msg += 'Check the attached traceback for more info.'
    elif ERROR_NAME[self.code] == 'FAILED_UPLOAD':
      msg += 'Unable to upload files.\n'
      msg += f'A full traceback of the error has been saved in {self.directory}.'
    else: 
      msg = 'Completed without errors.'

    return msg

  def update_document(self):
    """Update the request document with the results of the workflow.
    """

    if self.traceback != '':
      filepath = f"{self.directory}{os.sep}error_traceback.txt"
      with open(filepath, 'w') as fid:
        fid.write(self.traceback)
      self.output_files.append(filepath)

    # upload result files
    uploads = []
    for file in self.output_files:
      try: 
        file_obj = core.ELN.upload_file(open(file, 'rb'))
        uploads.append(file_obj)
      except:
        self.traceback += traceback.format_exc()
        self.code = ERROR_CODE['FAILED_UPLOAD']

    self.info += self.summary()
    print(self.info)

    # link result files to document
    fields = self.document['fields']
    # print([f['name'] for f in core.ELN.get_form(self.document['form']['id'])['fields']])#DEBUG
    # print([f['name'] for f in fields])#DEBUG
    for i in range(len(fields)):
      if self.field_name['completed'] is not None and (fields[i]['name']==self.field_name['completed']) and not self.code:
        fields[i]['content'] = 'yes'
      elif fields[i]['name']==self.field_name['output']:
        if self.code == ERROR_CODE['ALREADY_COMPLETED']: 
          fields[i]['content'] += '<p>'+self.info.replace('\n', '<br/>')+'</p>'
          continue
        else:
          fields[i]['content'] = '<p>'+self.info.replace('\n', '<br/>')+'</p>'
        for upload in uploads:
          fields[i]['content'] += f"<br/>{core.html_ref(upload)}"

    # update document
    # if self.code: return
    # print([f['name'] for f in fields])#DEBUG
    core.ELN.update_document(self.document['id'], fields=fields)

  def reset_document(self):
    """Reset the request document to an empty 'output' field and 'completed' field 'no'.
    """

    fields = self.document['fields']
    for i in range(len(fields)):
      if fields[i]['name']==self.field_name['completed'] and fields[i]['content'] == 'yes':
        fields[i]['content'] = 'no'
      elif fields[i]['name']==self.field_name['output']:
        fields[i]['content'] = ''

    core.ELN.update_document(self.document['id'], fields=fields)
    print(f"Reset Rspace document {self.document['id']}")

  def run(self):
    """Run the entire Workflow.

    This function should be redefined for every subclass of Workflow.
    """

    self.prepare()
    if not self.code:
      try: self.workflow()
      except:
        self.code = ERROR_CODE['FAILED_WORKFLOW']
        self.traceback += traceback.format_exc()
    self.update_document()
    